/*
 * Copyright © 2006, DoCoMo Communications Laboratories USA, Inc.,
 *   the DoCoMo SEND Project
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of DoCoMo Communications Laboratories USA, Inc., its
 *    parents, affiliates, subsidiaries, theDoCoMo SEND Project nor the names
 *    of the Project's contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL DOCOMO COMMUNICATIONS LABORATORIES USA,
 *  INC., ITS PARENTS, AFFILIATES, SUBSIDIARIES, THE PROJECT OR THE PROJECT'S
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/mbuf.h>
#include <sys/socket.h>

#include <arpa/inet.h>

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>

#include <netinet6/send.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "config.h"

#include "sendd_local.h"
#include "os_specific.h"
#include "snd_proto.h"
#include "dbg.h"

#include <applog.h>
#include <sbuff.h>

static int icmp6sock 	= -1;
#ifdef SND_OS_freebsd
static int sndsock	= -1;
#endif

#ifdef	DEBUG
static struct dlog_desc dbg = {
	.desc = "net",
	.ctx = SENDD_NAME
};
static char abuf[INET6_ADDRSTRLEN];
#endif

#ifdef SND_OS_freebsd
/* Per-interface info */
struct snd_ifinfo {
	struct list_head list;
	char		name[32];
	int		ifidx;
	int		snds;
};
static DEFINE_LIST_HEAD(ifaces);

/* Data packet meta data */
struct snd_packet_info {
	struct snd_ifinfo *ifinfo;
	int		in;
	int		ifidx;
};

void snd_sock_read(struct snd_ifinfo *p);
#endif

/* TODO: dynamically size according to MTU */
struct sbuff *
snd_get_buf(void)
{
	struct sbuff *b;

	if ((b = sbuff_alloc(SND_MAX_PKT)) == NULL) {
		APPLOG_NOMEM();
		return (NULL);
	}
	return (b);
}

#ifdef SND_MAX_PKT
void
snd_dispatch_fds(fd_set *fds)
{
	struct snd_ifinfo *p;

	list_for_each_entry(p, &ifaces, list) {
		if (FD_ISSET(sndsock, fds)) {
			snd_sock_read(p);
		}
	}
}

int
snd_handle_iface(const char *ifname, int ifidx)
{
	applog(LOG_ERR, "snd_handle_iface");
	struct snd_ifinfo *p;

	if ((p = malloc(sizeof (*p))) == NULL) {
		APPLOG_NOMEM();
		return (-1);
	}
	memset(p, 0, sizeof (*p));

	snprintf(p->name, sizeof (p->name), "%s", ifname);
	DBG(&dbg_snd, "snd_handle_iface -> p->name = %s", p->name);
	p->ifidx = ifidx;
	DBG(&dbg_snd, "snd_handle_iface -> p->ifidx = %d", ifidx);
	p->snds = -1;
	list_add_tail(&p->list, &ifaces);

	return (0);
}

void
snd_deliver_pkt(void *p, struct sbuff *b, int drop, int changed)
{
	struct snd_packet_info *pi;
	struct snd_hdr *snd_hdr;

	if (drop) {
		snd_put_buf(b);
		return;
	}

	pi = (struct snd_packet_info *)(b->head);
	b->data -= SEND_HDR_LEN;
	b->len += SEND_HDR_LEN;
	snd_hdr = sbuff_data(b);
	snd_hdr->direction = pi->in;
	snd_hdr->ifidx = pi->ifinfo->ifidx;

        if (send(sndsock, b->data, b->len, 0) < 0) {
                DBG(&dbg_snd, "Failed to send SEND message back to kernel.");
                perror("Failed");
                snd_put_buf(b);
                return;
        }

	snd_put_buf(b);

	return;
}


void
snd_add_fds(fd_set *fds, int *maxfd, int sock)
{
	struct snd_ifinfo *p;

	list_for_each_entry(p, &ifaces, list) {
		*maxfd = *maxfd > p->snds ? *maxfd : p->snds;
		p->snds = sock;
		FD_SET(p->snds, fds);
	}
}
#endif

int
freebsd_snd_init(void)
{

        if (list_empty(&ifaces)) {
                applog(LOG_ERR, "SEND must be active on at least one iface");
                return (-1);
        }

        return (0);
}

void
snd_put_buf(struct sbuff *b)
{
	sbuff_free(b);
}

int
snd_send_icmp(struct sbuff *b, struct sockaddr_in6 *sin, int ifidx)
{
	struct iovec iov = { b->head, b->len };
	struct in6_pktinfo *ipi;
	uint8_t cbuf[CMSG_SPACE(sizeof (*ipi))];
	struct cmsghdr *cmsg = (struct cmsghdr *)cbuf;
	struct msghdr msg = {
		.msg_name = sin,
		.msg_namelen = sizeof (*sin),
		.msg_iov = &iov,
		.msg_iovlen = 1,
		.msg_control = cbuf,
		.msg_controllen = sizeof (cbuf),
	};

	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = IPV6_PKTINFO;
	cmsg->cmsg_len = CMSG_LEN(sizeof (*ipi));

	ipi = (struct in6_pktinfo *)CMSG_DATA(cmsg);
	memset(ipi, 0, sizeof (*ipi));
	ipi->ipi6_ifindex = ifidx;

	DBG(&dbg, "sending %d bytes to %s (%d)", b->len,
	    inet_ntop(AF_INET6, &sin->sin6_addr, abuf, sizeof (abuf)),
	    ifidx);

	if (sendmsg(icmp6sock, &msg, 0) < 0) {
		DBG(&dbg_snd, "sendmsg: %s", strerror(errno));
		return (-1);
	}
	return (0);
}

#ifdef SND_OS_freebsd
void
snd_sock_read(struct snd_ifinfo *p)
{
	struct snd_hdr *snd_hdr;
	int direction;
	struct sbuff *b;
	struct snd_packet_info *pi;
	int n;

	if ((b = snd_get_buf()) == NULL) {
		goto done;
	}

	pi = sbuff_data(b);
	sbuff_advance(b, sizeof (*pi));

	n = recv(sndsock, b->data, b->rem, 0);
	if (n < 0) {
		applog(LOG_ERR, "%s: read: %s", __FUNCTION__, strerror(errno));
		goto done;
	} else
		DBG(&dbg_snd, "%d bytes received on routing socket. (%d)", n, b->rem);

	b->len = n;

	snd_hdr = sbuff_pull(b, SEND_HDR_LEN);
	direction = snd_hdr->direction;
	switch(direction) {
	case SND_IN:
		applog(LOG_ERR, "Direction: SND_IN");
		pi->ifinfo = p;
		pi->in = 1;
		snd_recv_pkt(b, p->ifidx, SND_IN, pi);
		break;
	case SND_OUT:
                applog(LOG_ERR, "Direction: SND_OUT");
                pi->ifinfo = p;
                pi->in = 0;
                snd_recv_pkt(b, p->ifidx, SND_OUT, pi);
                break;
	default:
		applog(LOG_ERR, "Unknown SEND pkt header: unknown direction.");
	}

done:
	/* ToDo: Free memory! */
	snd_put_buf(b);
}
#endif

/*
 * TODO: Linux is not yet up-to-date with rfc3542, specifically in that
 * it uses the socket option IPV6_PKTINFO instead of IPV6_RECVPKTINFO.
 * So for now we just use the older sin6_scope_id, which still works.
 */
void
snd_icmp_sock_read(void)
{
	struct sockaddr_in6 sin[1];
	struct sbuff *b;
	uint8_t *type;
	socklen_t slen;
	int r;

	if ((b = snd_get_buf()) == NULL) {
		return;
	}

	slen = sizeof (*sin);
	if ((r = recvfrom(icmp6sock, b->head, b->rem, 0, (void *)sin, &slen))
	    < 0) {
		applog(LOG_ERR, "%s: recvfrom: %s", __FUNCTION__,
		       strerror(errno));
		goto done;
	}
	b->len = r;

	DBG(&dbg, "%d bytes from %s on IF %d", r,
	    inet_ntop(AF_INET6, &sin->sin6_addr, abuf, sizeof (abuf)),
	    sin->sin6_scope_id);

	/* Is this OK? */
	if (IN6_IS_ADDR_LOOPBACK(&sin->sin6_addr)) {
		DBG(&dbg, "Dropping request from loopback");
		goto done;
	}

	type = sbuff_data(b);
	switch (*type) {
	case ICMP6_SND_CPS:
		snd_handle_cps(b, sin, sin->sin6_scope_id);
		break;
	case ICMP6_SND_CPA:
		snd_handle_cpa(b, sin);
		break;
	case ND_ROUTER_ADVERT:
		snd_process_ra(sbuff_data(b), r, sin->sin6_scope_id,
			       &sin->sin6_addr);
		break;
	default:
		DBG(&dbg_snd, "Unhandled ICMP6 type %d", *type);
		break;
	}

done:
	snd_put_buf(b);
}

int
#ifdef SND_OS_freebsd
snd_net_init(int *icmp6socket, int *sndsocket)
#else
snd_net_init(void)
#endif
{
	int v;
	struct icmp6_filter filter;
#ifdef	DEBUG
	struct dlog_desc *dbgs[] = {
		&dbg,
		NULL
	};

	if (snd_applog_register(dbgs) < 0) {
		return (-1);
	}
#endif
#ifdef SND_OS_freebsd
#define IPPROTO_SEND	259
#endif

#ifdef SND_OS_freebsd
	if ((*sndsocket = socket(PF_INET, SOCK_RAW, IPPROTO_SEND)) < 0) {	
		applog(LOG_ERR, "%s: socket: %s", __FUNCTION__,
			strerror(errno));
		return(-1);
	} else {
		applog(LOG_ERR, "%s: SEND socket created.", __FUNCTION__);
	}
#endif
	if ((*icmp6socket = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
		applog(LOG_ERR, "%s: socket: %s", __FUNCTION__,
		       strerror(errno));
		return (-1);
	}
#ifdef SND_OS_freebsd
	icmp6sock = *icmp6socket;
	sndsock = *sndsocket;
#endif
	v = 255;
	if (setsockopt(icmp6sock, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &v,
		       sizeof (v)) < 0) {
		applog(LOG_ERR, "%s: setsockopt(IPV6_UNICAST_HOPS): %s",
		       __FUNCTION__, strerror(errno));
		return (-1);
	}
	v = 255;
	if (setsockopt(icmp6sock, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &v,
	    sizeof (v)) < 0) {
		applog(LOG_ERR, "%s: setsockopt(IPV6_MULTICAST_HOPS): %s",
		       __FUNCTION__, strerror(errno));
		return (-1);
	}

	ICMP6_FILTER_SETBLOCKALL(&filter);
	ICMP6_FILTER_SETPASS(ICMP6_SND_CPS, &filter);
	ICMP6_FILTER_SETPASS(ICMP6_SND_CPA, &filter);
	ICMP6_FILTER_SETPASS(ND_ROUTER_ADVERT, &filter);

	if (setsockopt(icmp6sock, IPPROTO_ICMPV6, ICMP6_FILTER, &filter,
		       sizeof (filter)) < 0) {
		applog(LOG_ERR, "%s: setsockopt(ICMP6_FILTER): %s",
		       __FUNCTION__, strerror(errno));
		return (-1);
	}

#ifdef SND_OS_freebsd
	return (0);
#else
	return(icmp6sock);
#endif
}
